#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-21
# @Filename: unwind.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.kaiju import explode, unwind

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from jaeger.actor import JaegerActor
    from jaeger.fps import FPS


__all__ = ["unwind_command", "explode_command"]


@jaeger_parser.command(name="unwind")
@click.option(
    "--collision-buffer",
    type=click.FloatRange(1.6, 3.0),
    help="Custom collision buffer",
)
@click.option(
    "--force",
    is_flag=True,
    help="Execute unwind even in presence of deadlocks.",
)
@click.option(
    "--status",
    is_flag=True,
    help="Checks if the array is folded.",
)
async def unwind_command(
    command: Command[JaegerActor],
    fps: FPS,
    collision_buffer: float | None = None,
    force: bool = False,
    status: bool = False,
):
    """Sends the FPS to folded."""

    if fps.locked:
        command.fail(error="The FPS is locked.")

    await fps.update_position()
    positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.positioners.values()}

    positions_array = fps.get_positions()

    if len(positions_array) == 0:
        return command.fail("No positioners found.")

    if await fps.is_folded():
        command.info(folded=True)
        return command.finish("All positioners are folded.")
    else:
        command.info(folded=False)
        if status:
            return command.finish("The array is not folded.")

    command.debug(text="Calculating unwind trajectory.")

    try:
        trajectory = await unwind(
            positions,
            collision_buffer=collision_buffer,
            disabled=[pid for pid in fps.positioners if fps.positioners[pid].disabled],
            force=force,
        )
    except (ValueError, TrajectoryError) as err:
        return command.fail(error=f"Failed calculating trajectory: {err}")

    command.info("Executing unwind trajectory.")

    try:
        await fps.send_trajectory(trajectory, command=command)
    except TrajectoryError as err:
        return command.fail(error=f"Trajectory failed with error: {err}")

    command.finish()


@jaeger_parser.command(name="explode")
@click.option("--one", type=int, help="Only explode this positioner.")
@click.argument("EXPLODE-DEG", type=float)
async def explode_command(
    command: Command[JaegerActor],
    fps: FPS,
    explode_deg: float,
    one: int | None = None,
):
    """Explodes the FPS."""

    if fps.locked:
        command.fail(error="The FPS is locked.")

    command.debug(text="Calculating explode trajectory.")

    positions = {p.positioner_id: (p.alpha, p.beta) for p in fps.positioners.values()}

    try:
        trajectory = await explode(
            positions,
            explode_deg=explode_deg,
            disabled=[pid for pid in fps.positioners if fps.positioners[pid].disabled],
            positioner_id=one,
        )
    except (JaegerError, ValueError, TrajectoryError) as err:
        return command.fail(error=f"Failed calculating trajectory: {err}")

    command.info("Executing explode trajectory.")

    try:
        await fps.send_trajectory(trajectory, command=command)
    except TrajectoryError as err:
        return command.fail(error=f"Trajectory failed with error: {err}")

    command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-10
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib
import re

from typing import TYPE_CHECKING

import numpy
import pandas
from pydl.pydlutils.sdss import yanny

from coordio.conv import (
    positionerToTangent,
    tangentToPositioner,
    tangentToWok,
    wokToTangent,
)
from coordio.defaults import POSITIONER_HEIGHT, calibration, getHoleOrient

import jaeger.target
from jaeger import config, log
from jaeger.exceptions import JaegerError
from jaeger.kaiju import (
    decollide_in_executor,
    get_path_pair_in_executor,
    get_robot_grid,
)


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = [
    "wok_to_positioner",
    "positioner_to_wok",
    "copy_summary_file",
    "read_confSummary",
]


def wok_to_positioner(
    hole_id: str,
    site: str,
    fibre_type: str,
    xwok: float,
    ywok: float,
    zwok: float = POSITIONER_HEIGHT,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Converts from wok to positioner coordinates.

    Returns arrays with the positioner and tangent coordinates.

    """

    positioner_data = calibration.positionerTable.loc[(site, hole_id)]

    hole_orient = getHoleOrient(site, hole_id)

    if fibre_type == "APOGEE":
        xBeta = positioner_data.apX
        yBeta = positioner_data.apY
    elif fibre_type == "BOSS":
        xBeta = positioner_data.bossX
        yBeta = positioner_data.bossY
    elif fibre_type == "Metrology":
        xBeta = positioner_data.metX
        yBeta = positioner_data.metY
    else:
        raise ValueError(f"Invalid fibre type {fibre_type}.")

    tangent = wokToTangent(
        xwok,
        ywok,
        zwok,
        *hole_orient,
        dx=positioner_data.dx,
        dy=positioner_data.dy,
    )

    alpha, beta, _ = tangentToPositioner(
        tangent[0][0],
        tangent[1][0],
        xBeta,
        yBeta,
        la=positioner_data.alphaArmLen,
        alphaOffDeg=positioner_data.alphaOffset,
        betaOffDeg=positioner_data.betaOffset,
    )

    return (
        numpy.array([alpha, beta]),
        numpy.array([tangent[0][0], tangent[1][0], tangent[2][0]]),
    )


def positioner_to_wok(
    hole_id: str,
    site: str,
    fibre_type: str,
    alpha: float,
    beta: float,
):
    """Convert from positioner to wok coordinates.

    Returns xyz wok and tangent coordinates as a tuple of arrays.

    """

    positioner_data = calibration.positionerTable.loc[(site, hole_id)]
    wok_data = calibration.wokCoords.loc[(site, hole_id)]

    b = wok_data[["xWok", "yWok", "zWok"]]
    iHat = wok_data[["ix", "iy", "iz"]]
    jHat = wok_data[["jx", "jy", "jz"]]
    kHat = wok_data[["kx", "ky", "kz"]]

    if fibre_type == "APOGEE":
        xBeta = positioner_data.apX
        yBeta = positioner_data.apY
    elif fibre_type == "BOSS":
        xBeta = positioner_data.bossX
        yBeta = positioner_data.bossY
    elif fibre_type == "Metrology":
        xBeta = positioner_data.metX
        yBeta = positioner_data.metY
    else:
        raise ValueError(f"Invlid fibre type {fibre_type}.")

    tangent = positionerToTangent(
        alpha,
        beta,
        xBeta,
        yBeta,
        la=positioner_data.alphaArmLen,
        alphaOffDeg=positioner_data.alphaOffset,
        betaOffDeg=positioner_data.betaOffset,
    )

    wok = tangentToWok(
        tangent[0],
        tangent[1],
        0,
        b,
        iHat,
        jHat,
        kHat,
        dx=positioner_data.dx,
        dy=positioner_data.dy,
    )

    wok_coords = numpy.array(wok)

    return wok_coords, numpy.array([tangent[0], tangent[1], 0])


async def create_random_configuration(
    fps: FPS,
    seed: int | None = None,
    safe=True,
    uniform: tuple[float, ...] | None = None,
    collision_buffer: float | None = None,
    max_deadlocks: int = 6,
    deadlock_retries: int = 5,
    n_failed: int = 0,
    max_retries: int = 5,
    path_generation_mode: str | None = None,
    **kwargs,
):
    """Creates a random configuration using Kaiju."""

    from jaeger.target.configuration import ManualConfiguration

    seed = seed or numpy.random.randint(0, 1000000)
    numpy.random.seed(seed)

    robot_grid = get_robot_grid(fps, seed=seed, collision_buffer=collision_buffer)

    alphaL, betaL = config["kaiju"]["lattice_position"]

    # We use Kaiju for convenience in the non-safe mode.
    for robot in robot_grid.robotDict.values():

        if robot.isOffline:
            continue

        if uniform is not None:
            alpha0, alpha1, beta0, beta1 = uniform
            robot.setAlphaBeta(
                numpy.random.uniform(alpha0, alpha1),
                numpy.random.uniform(beta0, beta1),
            )

        else:
            if safe:
                safe_mode = config["safe_mode"]
                if isinstance(safe_mode, bool) or safe_mode is None:
                    safe_mode = {"min_beta": 165, "max_beta": 175}

                robot.setAlphaBeta(
                    numpy.random.uniform(0, 359.9),
                    numpy.random.uniform(
                        safe_mode["min_beta"],
                        175.0,
                    ),
                )

            else:
                robot.setXYUniform()

        robot.setDestinationAlphaBeta(alphaL, betaL)

    # Confirm that the configuration is valid. This should only matter
    # for full range random configurations.
    try:
        robot_grid, _ = await decollide_in_executor(robot_grid, simple=True)
        grid_data = {
            robot.id: (robot.alpha, robot.beta)
            for robot in robot_grid.robotDict.values()
        }
    except JaegerError:
        raise JaegerError("Decollision failed. Cannot create random configuration.")

    _, _, did_fail, deadlocks = await get_path_pair_in_executor(
        robot_grid,
        path_generation_mode=path_generation_mode,
    )

    # If too many deadlocks, just try a new seed.
    n_deadlock = len(deadlocks)
    if did_fail and n_deadlock > max_deadlocks:
        if n_failed >= max_retries:
            raise JaegerError("Reached the limit of retries.")

        log.warning("Too many deadlocked robots. Trying new seed.")
        return await create_random_configuration(
            fps,
            safe=safe,
            uniform=uniform,
            collision_buffer=collision_buffer,
            deadlock_retries=deadlock_retries,
            n_failed=n_failed + 1,
            path_generation_mode=path_generation_mode,
        )

    if did_fail and n_deadlock > 0:
        # Now the fun part, if there are only a few deadlocks, try assigning them
        # a random position.
        log.warning(f"Found {n_deadlock} deadlocked robots. Trying to unlock.")
        for nn in range(1, deadlock_retries + 1):
            log.info(f"Retry {nn} out of {deadlock_retries}.")

            to_replace_robot = numpy.random.choice(deadlocks)

            robot_grid = get_robot_grid(
                fps,
                seed=seed + 1,
                collision_buffer=collision_buffer,
            )

            for robot in robot_grid.robotDict.values():

                if robot.isOffline:
                    continue

                if robot.id == to_replace_robot:
                    robot.setXYUniform()
                else:
                    robot.setAlphaBeta(*grid_data[robot.id])

            try:
                robot_grid, _ = await decollide_in_executor(robot_grid, simple=True)
                grid_data = {
                    robot.id: (robot.alpha, robot.beta)
                    for robot in robot_grid.robotDict.values()
                }
            except JaegerError:
                raise JaegerError(
                    "Failed creating random configuration: cannot remove deadlocks."
                )

            _, _, did_fail, deadlocks = await get_path_pair_in_executor(
                robot_grid,
                path_generation_mode=path_generation_mode,
            )
            if did_fail is False:
                log.info("Random configuration has been unlocked.")
                break
            else:
                log.info(f"{len(deadlocks)} deadlocks remaining.")

            if nn == deadlock_retries:
                log.warning("Failed unlocking. Trying new seed.")
                return await create_random_configuration(
                    fps,
                    seed=seed + 1,
                    safe=safe,
                    uniform=uniform,
                    collision_buffer=collision_buffer,
                    deadlock_retries=deadlock_retries,
                    path_generation_mode=path_generation_mode,
                )

    pT = calibration.positionerTable.copy().reset_index()

    # Build an assignment dictionary.
    data = {}
    for pid in grid_data:
        holeID = pT.loc[pT.positionerID == pid].holeID.values[0]
        data[holeID] = {
            "alpha": grid_data[pid][0],
            "beta": grid_data[pid][1],
            "fibre_type": "Metrology",
        }

    return ManualConfiguration(data, **kwargs)


def copy_summary_file(
    configuration_id0: int,
    configuration_id1: int,
    design_id1: int | None = None,
    flavour: str = "",
):
    """Copies a summary file optionally modifying its design_id."""

    orig_file = jaeger.target.Configuration._get_summary_file_path(
        configuration_id0,
        config["observatory"],
        flavour,
    )

    if not os.path.exists(orig_file):
        return

    new_file = jaeger.target.Configuration._get_summary_file_path(
        configuration_id1,
        config["observatory"],
        flavour,
    )

    new_path = pathlib.Path(new_file)
    new_path.parent.mkdir(parents=True, exist_ok=True)

    summary_data = open(orig_file, "r").read()
    summary_data = re.sub(
        r"(configuration_id\s)[0-9]+",
        rf"\g<1>{configuration_id1}",
        summary_data,
    )

    if design_id1:
        summary_data = re.sub(
            rf"confSummary(F?)\-{configuration_id0}\.par",
            rf"confSummary\g<1>-{configuration_id1}.par",
            summary_data,
        )

        summary_data = re.sub(
            r"(design_id\s)[0-9]+",
            rf"\g<1>{design_id1}",
            summary_data,
        )

        summary_data = re.sub(
            r"cloned_from -999",
            f"cloned_from {configuration_id0}",
            summary_data,
        )

    with open(new_path, "w") as f:
        f.write(summary_data)


def read_confSummary(input: str | pathlib.Path | int, flavour: str = "") -> tuple:
    """Reads a configuration summary file and returns the header and data frame."""

    if isinstance(input, (str, pathlib.Path)):
        path = input
    else:
        sdsscore_dir = pathlib.Path(os.environ["SDSSCORE_DIR"])
        summary_files = sdsscore_dir / "apo" / "summary_files"
        conf_xx = summary_files / f"{int(input/100):04d}XX"
        path = conf_xx / f"confSummary{flavour}-{input}.par"

    if not os.path.exists(str(path)):
        raise FileNotFoundError(f"{path} does not exist.")

    y = yanny(str(path))
    header = dict(y)

    fibermap = header.pop("FIBERMAP")
    fibermap = fibermap[[col for col in fibermap.dtype.names if col != "mag"]]

    df = pandas.DataFrame(fibermap)

    for col in df.select_dtypes("object").columns:
        df[col] = df[col].str.decode("utf-8")

    for key, value in header.items():
        try:
            header[key] = int(value)
        except ValueError:
            try:
                header[key] = float(value)
            except ValueError:
                pass

    return header, df.set_index(["positionerId", "fiberType"])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['jaeger',
 'jaeger.actor',
 'jaeger.actor.commands',
 'jaeger.commands',
 'jaeger.interfaces',
 'jaeger.scripts',
 'jaeger.target',
 'jaeger.utils']

package_data = \
{'': ['*'], 'jaeger': ['etc/*', 'etc/sextants/*']}

install_requires = \
['astropy>=5.0.0,<6.0.0',
 'click_default_group>=1.2.2,<2.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.4,<2.0.0',
 'progressbar2>=3.39.3,<4.0.0',
 'pydl==1.0.0rc1',
 'sdss-clu>=1.5.7,<2.0.0',
 'sdss-coordio>=1.3.1,<2.0.0',
 'sdss-drift>=0.4.1,<0.5.0',
 'sdss-kaiju>=1.3.0,<2.0.0',
 'sdssdb>=0.5.5,<0.6.0',
 'sdsstools>=0.5.0,<0.6.0',
 'sep>=1.2.0,<2.0.0',
 'simplification==0.5.22',
 'tables>=3.6.1,<4.0.0',
 'zc.lockfile>=2.0,<3.0']

entry_points = \
{'console_scripts': ['jaeger = jaeger.__main__:jaeger',
                     'robotcalib = jaeger.scripts.robotcalib:robotcalib']}

setup_kwargs = {
    'name': 'jaeger',
    'version': '1.1.1a0',
    'description': 'Controllers for the SDSS-V FPS',
    'long_description': '# jaeger\n\n![Versions](https://img.shields.io/badge/python->3.8-blue)\n[![Documentation Status](https://readthedocs.org/projects/jaeger/badge/?version=latest)](https://sdss-jaeger.readthedocs.io/en/latest/?badge=latest)\n[![Build](https://img.shields.io/github/workflow/status/sdss/jaeger/Test)](https://github.com/sdss/jaeger/actions)\n[![codecov](https://codecov.io/gh/sdss/jaeger/branch/main/graph/badge.svg)](https://codecov.io/gh/sdss/jaeger)\n\n[jaeger](http://pacificrim.wikia.com/wiki/Jaeger>) provides high level control for the SDSS-V [Focal Plane System](https://wiki.sdss.org/display/FPS). Some of the features that jaeger provide are:\n\n- Wraps the low level CAN commands for simpler use.\n- Provides a framework that is independent of the CAN interface used (by using the [python-can](https://python-can.readthedocs.io/en/master/) library).\n- Interfaces with [kaiju](https://github.com/sdss/kaiju) to provide anticollision path planning for trajectories.\n- Implements status and position update loops.\n- Provides implementations for commonly used tasks (e.g., go to position, send trajectory).\n- Interfaces with the Instrument Electronics Box modbus PLC controller.\n- Provides a TCP/IP interface to send commands and output keywords using the SDSS-standard formatting.\n\nThe code for jaeger is developed in [GitHub](https://github.com/sdss/jaeger) and can be installed using [sdss_install](https://github.com/sdss/sdss_install) or by running\n\n```console\npip install --upgrade sdss-jaeger\n```\n\nTo check out the development version do\n\n```console\ngit clone https://github.com/sdss/jaeger.git\n```\n\njaeger is developed as an [asyncio](https://docs.python.org/3/library/asyncio.html) library and a certain familiarity with asynchronous programming is required. The actor functionality (TCP/IP connection, command parser, inter-actor communication) is built on top of [CLU](https://github.com/sdss/clu).\n\n## A simple jaeger program\n\n```python\nimport asyncio\nfrom jaeger import FPS, log\n\nasync def main():\n\n    # Set logging level to DEBUG\n    log.set_level(0)\n\n    # Initialise the FPS instance.\n    fps = FPS()\n    await fps.initialise()\n\n    # Print the status of positioner 4\n    print(fps[4].status)\n\n    # Send positioner 4 to alpha=90, beta=45\n    await pos.goto(alpha=90, beta=45)\n\n    # Cleanly finish all pending tasks and exit\n    await fps.shutdown()\n\nasyncio.run(main())\n```\n',
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sdss/jaeger',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-09
# @Filename: bootloader.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import contextlib
import os
import pathlib
import time
import warnings
import zlib

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import numpy

from jaeger import can_log, config, log
from jaeger.commands import Command, CommandID
from jaeger.exceptions import JaegerError, JaegerUserWarning
from jaeger.maskbits import BootloaderStatus
from jaeger.utils import int_to_bytes


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = [
    "load_firmware",
    "StartFirmwareUpgrade",
    "GetFirmwareVersion",
    "SendFirmwareData",
]


async def load_firmware(
    fps: FPS,
    firmware_file: str | pathlib.Path,
    positioners: Optional[List[int]] = None,
    messages_per_positioner: Optional[int] = None,
    force: bool = False,
    show_progressbar: bool = False,
    progress_callback: Optional[Callable[[int, int], Any]] = None,
    stop_logging: bool = True,
):
    """Convenience function to run through the steps of loading a new firmware.

    This function is a coroutine and not intendend for direct use. Use the
    ``jaeger`` CLI instead.

    Parameters
    ----------
    fps
        `~jaeger.fps.FPS` instance to which the commands will be sent.
    firmware_file
        Binary file containing the firmware to load.
    positioners
        A list of positioner ids whose firmware to update, or `None` to update
        all the positioners in ``fps``.
    messages_per_positioner
        How many messages to send to each positioner at once. This can improve the
        performance but also overflow the CAN bus buffer. With the default value of
        `None`, reverts to the configuration value
        ``positioner.firmware_messages_per_positioner``.
    force
        Forces the firmware load to continue even if some positioners are not
        responding or are not in bootloader mode.
    show_progressbar
        Whether to show a progress bar.
    progress_callback
        A function to call as data gets transferred to the positioners. The
        callback is called with ``(current_chunk, n_chuck)`` where
        ``current_chunk`` is the number of the data chunk being sent and
        ``n_chunk`` is the total number of chunks in the data package.
    stop_logging
        Disable logging to file for the CAN logger to improve performance.

    """

    if show_progressbar:
        try:
            import progressbar
        except ImportError:
            warnings.warn(
                "progressbar2 is not installed. Cannot show a progress bar.",
                JaegerUserWarning,
            )
            progressbar = None
            show_progressbar = False
    else:
        progressbar = None

    start_time = time.time()

    firmware_file = pathlib.Path(firmware_file)
    assert firmware_file.exists(), "cannot find firmware file"

    log.info(f"firmware file {firmware_file!s} found.")

    # Open firmware data as binary.
    firmware_data = open(firmware_file, "rb")
    crc32 = zlib.crc32(firmware_data.read())
    filesize = os.path.getsize(firmware_file)

    # Check to make sure all positioners are in bootloader mode.
    valid_positioners = []
    n_bad = 0

    for positioner_id in fps.positioners:

        if positioners is not None and positioner_id not in positioners:
            continue

        positioner = fps.positioners[positioner_id]

        if (
            not positioner.is_bootloader()
            or BootloaderStatus.BOOTLOADER_INIT not in positioner.status
            or BootloaderStatus.UNKNOWN in positioner.status
        ):

            n_bad += 1
            continue

        valid_positioners.append(positioner)

    if len(valid_positioners) == 0:
        raise JaegerError(
            "no positioners found in bootloader mode or with valid status."
        )

    if n_bad > 0:
        msg = f"{n_bad} positioners not in bootloader mode or state is invalid."
        if force:
            warnings.warn(msg + " Proceeding becasuse force=True.", JaegerUserWarning)
        else:
            raise JaegerError(msg)

    log.info("stopping pollers")
    await fps.pollers.stop()

    log.info(f"upgrading firmware on {len(valid_positioners)} positioners.")

    start_firmware_payload = int_to_bytes(filesize) + int_to_bytes(crc32)

    log.info(f"CRC32: {crc32}")
    log.info(f"File size: {filesize} bytes")

    pids = [pos.positioner_id for pos in valid_positioners]

    cmd = await fps.send_command(
        CommandID.START_FIRMWARE_UPGRADE,
        positioner_ids=pids,
        data=[start_firmware_payload],
    )

    if cmd.status.failed or cmd.status.timed_out:
        log.error("firmware upgrade failed.")
        return False

    # Restore pointer to start of file
    firmware_data.seek(0)

    log.info("starting data send.")

    if stop_logging and can_log.fh:
        fh_handler = can_log.handlers.pop(can_log.handlers.index(can_log.fh))
    else:
        fh_handler = None

    chunk_size = 8
    n_chunks = int(numpy.ceil(filesize / chunk_size))

    with contextlib.ExitStack() as stack:

        if show_progressbar and progressbar:
            bar = stack.enter_context(progressbar.ProgressBar(max_value=n_chunks))
        else:
            bar = None

        messages_default = config["positioner"]["firmware_messages_per_positioner"]
        messages_per_positioner = messages_per_positioner or messages_default
        assert isinstance(messages_per_positioner, int)

        ii = 0
        while True:
            cmds = []
            stop = False
            for __ in range(messages_per_positioner):
                chunk = firmware_data.read(chunk_size)
                packetdata = bytearray(chunk)
                # packetdata.reverse()  # IMPORTANT! no longer needed for P1

                if len(packetdata) == 0:
                    stop = True
                    break

                cmds.append(
                    fps.send_command(
                        CommandID.SEND_FIRMWARE_DATA,
                        positioner_ids=pids,
                        data=[packetdata],
                        timeout=15,
                    )
                )

            await asyncio.gather(*cmds)

            if any(cmd.status.failed or cmd.status.timed_out for cmd in cmds):
                log.error("firmware upgrade failed.")
                if fh_handler:
                    can_log.addHandler(fh_handler)
                return False

            ii += messages_per_positioner

            if show_progressbar and bar:
                if ii < n_chunks:
                    bar.update(ii)

            if progress_callback:
                progress_callback(ii, n_chunks)

            if stop:
                break

    log.info("firmware upgrade complete.")

    if fh_handler:
        can_log.addHandler(fh_handler)

    total_time = time.time() - start_time
    log.info(f"upgrading firmware took {total_time:.2f} s.")

    return True


class GetFirmwareVersion(Command):

    command_id = CommandID.GET_FIRMWARE_VERSION
    broadcastable = True
    safe = True
    bootloader = True

    def get_replies(self) -> Dict[int, Any]:
        return self.get_firmware()

    def get_firmware(self, positioner_id=None) -> Dict[int, str]:
        """Returns the firmware version string.

        Parameters
        ----------
        positioner_id : int
            The positioner for which to return the version. If `None` returns
            a dictionary with the firmware version of all the positioners that
            replied.

        Returns
        -------
        firmware
            A string or dictionary of string with the firmware version(s),
            with the format ``'XX.YY.ZZ'`` where ``YY='80'`` if the positioner
            is in bootloader mode.

        Raises
        ------
        ValueError
            If no positioner with ``positioner_id`` has replied.

        """

        def format_version(reply):
            return ".".join(format(byt, "02d") for byt in reply.data[0:3][::-1])

        firmwares = {}
        for reply in self.replies:
            version = format_version(reply)
            firmwares[reply.positioner_id] = version

        return firmwares

    @staticmethod
    def encode(firmware):
        """Returns the bytearray encoding the firmware version."""

        chunks = firmware.split(".")[::-1]

        data = b""
        for chunk in chunks:
            data += int_to_bytes(int(chunk), "u1")

        return data


class StartFirmwareUpgrade(Command):

    command_id = CommandID.START_FIRMWARE_UPGRADE
    broadcastable = False
    safe = True
    bootloader = True


class SendFirmwareData(Command):

    command_id = CommandID.SEND_FIRMWARE_DATA
    broadcastable = False
    safe = True
    bootloader = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-18
# @Filename: pollers.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from . import jaeger_parser


__all__ = ["pollers"]


@jaeger_parser.group()
def pollers():
    """Handle the positioner pollers."""

    pass


@pollers.command()
async def list(command, fps):
    """Lists available pollers."""

    poller_status = []
    for name in fps.pollers.names:
        poller_status.append(name + ("*" if fps.pollers[name].running else ""))

    command.finish(text=",".join(poller_status))


@pollers.command()
@click.argument("POLLER", type=str, required=False)
async def stop(command, fps, poller):
    """Stop pollers."""

    if poller is None:
        await fps.pollers.stop()
    else:
        if poller not in fps.pollers.names:
            command.fail(error="poller not found.")
        await fps.pollers[poller].stop()

    command.finish(text="pollers stopped")


@pollers.command()
@click.argument("POLLER", type=str, required=False)
async def start(command, fps, poller):
    """Start pollers."""

    if poller is None:
        fps.pollers.start()
    else:
        if poller not in fps.pollers.names:
            command.fail(error="poller not found.")
        fps.pollers[poller].start()

    command.finish(text="pollers started")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from typing import TYPE_CHECKING

from clu import Command
from clu.parsers.click import command_parser

from jaeger.actor import JaegerActor


jaeger_parser = command_parser

JaegerCommandType = Command[JaegerActor]


from .alerts import *
from .can import *
from .chiller import *
from .configuration import *
from .debug import *
from .disable import *
from .fvc import *
from .home import *
from .ieb import *
from .pollers import *
from .positioner import *
from .power import *
from .snapshot import *
from .talk import *
from .testing import *
from .unwind import *
from .version import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-10-08
# @Filename: trajectory.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import os
import pathlib
import time
import warnings
from glob import glob

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, cast

import numpy
from astropy.time import Time

import drift
from sdsstools import read_yaml_file

from jaeger import config, log
from jaeger.commands import Command, CommandID
from jaeger.exceptions import FPSLockedError, JaegerUserWarning, TrajectoryError
from jaeger.ieb import IEB
from jaeger.maskbits import FPSStatus, ResponseCode
from jaeger.utils import int_to_bytes


if TYPE_CHECKING:
    from clu.command import Command as CluCommand

    from jaeger import FPS
    from jaeger.actor import JaegerActor


__all__ = [
    "send_trajectory",
    "SendNewTrajectory",
    "SendTrajectoryData",
    "TrajectoryDataEnd",
    "SendTrajectoryAbort",
    "StartTrajectory",
    "StopTrajectory",
    "Trajectory",
]


MOTOR_STEPS = config["positioner"]["motor_steps"]
TIME_STEP = config["positioner"]["time_step"]


TrajectoryDataType = Dict[int, Dict[str, List[Tuple[float, float]]]]


async def send_trajectory(
    fps: FPS,
    trajectories: str | pathlib.Path | TrajectoryDataType,
    use_sync_line: bool | None = None,
    send_trajectory: bool = True,
    start_trajectory: bool = True,
    command: Optional[CluCommand[JaegerActor]] = None,
    dump: bool | str = True,
    extra_dump_data: dict[str, Any] = {},
) -> Trajectory:
    """Sends a set of trajectories to the positioners.

    This is a low-level function that raises errors when a problem is
    encountered. Most users should use `.FPS.send_trajectory` instead.
    `.send_trajectory` automates `.Trajectory` by calling the different
    methods in order, but provides less control.

    Parameters
    ----------
    fps
        The instance of `.FPS` that will receive the trajectory.
    trajectories
        Either a path to a YAML file to read or a dictionary with the
        trajectories. In either case the format must be a dictionary in
        which the keys are the ``positioner_ids`` and each value is a
        dictionary containing two keys: ``alpha`` and ``beta``, each
        pointing to a list of tuples ``(position, time)``, where
        ``position`` is in degrees and ``time`` is in seconds.
    use_sync_line
        If `True`, the SYNC line will be used to synchronise the beginning of
        all trajectories. Otherwise a `.START_TRAJECTORY` command will be sent
        over the CAN network. If `None`, defaults to the configuration parameter.
    send_trajectory
        If `True`, sends the trajectory to the positioners and returns the
        `.Trajectory` instance.
    start_trajectory
        If `True`, runs the trajectory after sending it. Otherwise, returns
        the `.Trajectory` instance after sending the data. Ignored if
        `send_trajectory=False`.
    command
        A command to which to output informational messages.
    dump
        Whether to dump a JSON with the trajectory to disk. If `True`,
        the trajectory is stored in ``/data/logs/jaeger/trajectories/<MJD>`` with
        a unique sequence for each trajectory. A string can be passed with a
        custom path. The dump file is created only if `.Trajectory.start` is
        called, regardless of whether the trajectory succeeds.
    extra_dump_data
        A dictionary with additional parameters to add to the dump JSON.

    Raises
    ------
    TrajectoryError
        If encounters a problem sending the trajectory.
    FPSLockedError
        If the FPS is locked.

    Examples
    --------
    ::

        >>> fps = FPS()
        >>> await fps.initialise()

        # Send a trajectory with two points for positioner 4
        >>> await fps.send_trajectory({4: {'alpha': [(90, 0), (91, 3)],
                                           'beta': [(20, 0), (23, 4)]}})

    """

    traj = Trajectory(
        fps,
        trajectories,
        dump=dump,
        extra_dump_data=extra_dump_data,
    )

    if use_sync_line is None:
        use_sync_line = config["fps"]["use_sync_line"]

    assert isinstance(use_sync_line, bool)

    if use_sync_line:
        if not isinstance(fps.ieb, IEB) or fps.ieb.disabled:
            raise TrajectoryError("IEB is not connected. Cannot use SYNC line.", traj)
        if (await fps.ieb.get_device("sync").read())[0] == "closed":
            raise TrajectoryError("The SYNC line is on high.", traj)

    if send_trajectory is False:
        return traj

    msg = "Sending trajectory data."
    log.debug(msg)
    if command:
        command.debug(msg)

    try:
        await traj.send()
    except TrajectoryError as err:
        raise TrajectoryError(
            f"Something went wrong sending the trajectory: {err}",
            err.trajectory,
        )
    finally:
        if traj.dump_file:
            traj.dump_trajectory()
            if command:
                command.debug(trajectory_dump_file=traj.dump_file)

    msg = f"Trajectory sent in {traj.data_send_time:.1f} seconds."
    log.debug(msg)
    if command:
        command.debug(msg)

    log.info(f"Expected time to complete trajectory: {traj.move_time:.2f} seconds.")
    if command and traj.move_time:
        command.info(move_time=round(traj.move_time, 2))

    if start_trajectory is False:
        return traj

    msg = "Starting trajectory ..."
    log.info(msg)
    if command:
        command.info(msg)

    try:
        await traj.start(use_sync_line=use_sync_line)
    except TrajectoryError as err:

        if traj.start_time is not None and traj.end_time is not None:
            elapsed = traj.end_time - traj.start_time
            elapsed_msg = f" Trajectory failed {elapsed:.2f} seconds after start."
        else:
            elapsed_msg = ""

        raise TrajectoryError(
            f"Something went wrong during the trajectory: {err}.{elapsed_msg}",
            err.trajectory,
        )
    finally:
        if traj.dump_file and command:
            command.debug(trajectory_dump_file=traj.dump_file)

    if command:
        command.info(folded=(await fps.is_folded()))

    msg = "All positioners have reached their destinations."
    log.info(msg)
    if command:
        command.info(msg)

    return traj


class Trajectory(object):
    """Prepares and sends a trajectory to the FPS.

    Most user will prefer using `.FPS.send_trajectory`, which automates the
    process. This class provides fine-grain control over the process.

    Parameters
    ----------
    fps
        The instance of `.FPS` that will receive the trajectory.
    trajectories
        Either a path to a YAML file to read or a dictionary with the
        trajectories. In either case the format must be a dictionary in
        which the keys are the ``positioner_ids`` and each value is a
        dictionary containing two keys: ``alpha`` and ``beta``, each
        pointing to a list of tuples ``(position, time)``, where
        ``position`` is in degrees and ``time`` is in seconds.
    dump
        Whether to dump a JSON with the trajectory to disk. If `True`,
        the trajectory is stored in ``/data/logs/jaeger/trajectories/<MJD>`` with
        a unique sequence for each trajectory. A string can be passed with a
        custom path. The dump file is created only if `.Trajectory.start` is
        called, regardless of whether the trajectory succeeds.
    extra_dump_data
        A dictionary with additional parameters to add to the dump JSON.

    Raises
    ------
    TrajectoryError
        If encounters a problem sending the trajectory.
    FPSLockedError
        If the FPS is locked.

    Examples
    --------
    Given the following two-point trajectory for positioner 4 ::

        points = {4: {'alpha': [(90, 0), (91, 3)],
                      'beta': [(20, 0), (23, 4)]}}

    the normal process to execute the trajectory is ::

        trajectory = Trajectory(fps, points)
        await trajectory.send()    # Sends the trajectory but does not yet execute it.
        await trajectory.start()   # This starts the trajectory (positioners move).

    """

    def __init__(
        self,
        fps: FPS,
        trajectories: str | pathlib.Path | TrajectoryDataType,
        dump: bool | str = True,
        extra_dump_data: dict[str, Any] = {},
    ):

        self.fps = fps
        self.trajectories: TrajectoryDataType

        if self.fps.locked:
            raise FPSLockedError(
                f"FPS is locked by {fps.locked_by}. Cannot send trajectories."
            )

        if self.fps.moving:
            raise TrajectoryError(
                "The FPS is moving. Cannot send new trajectory.",
                self,
            )

        if isinstance(trajectories, (str, pathlib.Path)):
            path = pathlib.Path(trajectories)
            if path.suffix == ".json":
                self.trajectories = json.loads(open(path, "r").read())["trajectory"]
            else:
                self.trajectories = cast(dict, read_yaml_file(trajectories))
        elif isinstance(trajectories, dict):
            self.trajectories = trajectories
        else:
            raise TrajectoryError("invalid trajectory data.", self)

        # List of positioners that failed receiving the trajectory and reason.
        self.failed_positioners: dict[int, str] = {}

        self.validate()

        #: Number of points sent to each positioner as a tuple ``(alpha, beta)``.
        self.n_points = {}

        #: The time required to complete the trajectory.
        self.move_time: float | None = None

        #: How long it took to send the trajectory.
        self.data_send_time: float | None = None

        self.failed = False
        self.send_new_trajectory_failed = False

        # Commands that will be sent. Mostly for inspection if the trajectory fails.
        self.data_send_cmd: Command | None = None
        self.end_traj_cmds: Command | None = None

        self.start_time: float | None = None
        self.end_time: float | None = None

        self.use_sync_line: bool = True
        self._ready_to_start = False

        self.dump_data = {
            "start_time": time.time(),
            "success": False,
            "trajectory_send_time": None,
            "trajectory_start_time": None,
            "end_time": None,
            "use_sync_line": True,
            "extra": extra_dump_data,
            "trajectory": self.trajectories,
            "initial_positions": self.fps.get_positions_dict(),
            "final_positions": {},
        }

        if dump is False:
            self.dump_file = None
        elif isinstance(dump, (str, pathlib.Path)):
            self.dump_file = str(dump)
        else:
            dirname = config["positioner"]["trajectory_dump_path"]
            mjd = str(int(Time.now().mjd))
            dirname = os.path.join(dirname, mjd)

            files = list(sorted(glob(os.path.join(dirname, "*.json"))))
            if len(files) == 0:
                seq = 1
            else:
                seq = int(files[-1].split("-")[-1].split(".")[0]) + 1

            self.dump_file = os.path.join(dirname, f"trajectory-{mjd}-{seq:04d}.json")

    def validate(self):
        """Validates the trajectory."""

        if len(self.trajectories) == 0:
            raise TrajectoryError("trajectory is empty.", self)

        if len(self.trajectories) != len(numpy.unique(list(self.trajectories))):
            raise TrajectoryError("Duplicate positioner trajectories.", self)

        for pid in self.trajectories:
            trajectory = self.trajectories[pid]

            if "alpha" not in trajectory or "beta" not in trajectory:
                self.failed_positioners[pid] = "NO_DATA"
                raise TrajectoryError(
                    f"Positioner {pid} missing alpha or beta data.",
                    self,
                )

            for arm in ["alpha", "beta"]:
                data = numpy.array(list(zip(*trajectory[arm]))[0])

                if arm == "beta":
                    if config.get("safe_mode", False):
                        if config["safe_mode"] is True:
                            min_beta = 160
                        else:
                            min_beta = config["safe_mode"]["min_beta"]
                        if numpy.any(data < min_beta):
                            self.failed_positioners[pid] = "SAFE_MODE"
                            raise TrajectoryError(
                                f"Positioner {pid}: safe mode is on "
                                f"and beta < {min_beta}.",
                                self,
                            )

    async def send(self):
        """Sends the trajectory but does not start it."""

        self.move_time = 0.0

        await self.fps.stop_trajectory()
        await self.fps.stop_trajectory(clear_flags=True)

        if not await self.fps.update_status(
            positioner_ids=list(self.trajectories),
            timeout=1.0,
        ):
            self.failed = True
            raise TrajectoryError("Some positioners did not respond.", self)

        # Check that all positioners are ready to receive a new trajectory.
        for pos_id in self.trajectories:

            positioner = self.fps.positioners[pos_id]
            status = positioner.status

            if positioner.disabled:
                raise TrajectoryError(
                    f"positioner_id={pos_id} is disabled/offline but was "
                    "included in the trajectory.",
                    self,
                )

            if (
                positioner.flags.DATUM_ALPHA_INITIALIZED not in status
                or positioner.flags.DATUM_BETA_INITIALIZED not in status
                or positioner.flags.DISPLACEMENT_COMPLETED not in status
            ):
                self.failed = True
                self.failed_positioners[pos_id] = "NOT_READY"
                raise TrajectoryError(
                    f"positioner_id={pos_id} is not ready to receive a trajectory.",
                    self,
                )

            traj_pos = self.trajectories[pos_id]

            self.n_points[pos_id] = (len(traj_pos["alpha"]), len(traj_pos["beta"]))

            # Gets maximum time for this positioner
            max_time_pos = max(
                [
                    max(list(zip(*traj_pos["alpha"]))[1]),
                    max(list(zip(*traj_pos["beta"]))[1]),
                ]
            )

            # Updates the global trajectory max time.
            if max_time_pos > self.move_time:
                self.move_time = max_time_pos

        new_traj_data = {}
        for pos_id in self.trajectories:
            data = SendNewTrajectory.get_data(
                self.n_points[pos_id][0],
                self.n_points[pos_id][1],
            )
            new_traj_data[pos_id] = data

        # Starts trajectory
        new_traj_cmd = await self.fps.send_command(
            "SEND_NEW_TRAJECTORY",
            positioner_ids=list(self.trajectories),
            data=new_traj_data,
        )

        if new_traj_cmd.status.failed or new_traj_cmd.status.timed_out:
            self.failed = True
            raise TrajectoryError("Failed sending SEND_NEW_TRAJECTORY.", self)

        start_trajectory_send_time = time.time()

        # How many points from the trajectory are we putting in each command.
        n_chunk = config["positioner"]["trajectory_data_n_points"]

        # Gets the maximum number of points for each arm for all positioners.
        max_points = numpy.max(list(self.n_points.values()), axis=0)
        max_points = {"alpha": max_points[0], "beta": max_points[1]}

        # Send chunks of size n_chunk to all the positioners in parallel.
        # Do alpha first, then beta.
        for arm in ["alpha", "beta"]:

            for jj in range(0, max_points[arm], n_chunk):

                data = {}
                send_trajectory_pids = []

                for pos_id in self.trajectories:

                    arm_chunk = self.trajectories[pos_id][arm][jj : jj + n_chunk]
                    if len(arm_chunk) == 0:
                        continue

                    send_trajectory_pids.append(pos_id)

                    positions = numpy.array(arm_chunk).astype(numpy.float64)
                    data_pos = SendTrajectoryData.calculate_positions(positions)

                    data[pos_id] = data_pos

                self.data_send_cmd = await self.fps.send_command(
                    "SEND_TRAJECTORY_DATA",
                    positioner_ids=send_trajectory_pids,
                    data=data,
                )

                status = self.data_send_cmd.status
                if status.failed or status.timed_out:
                    for reply in self.data_send_cmd.replies:
                        if reply.response_code != ResponseCode.COMMAND_ACCEPTED:
                            code = reply.response_code.name
                            self.failed_positioners[reply.positioner_id] = code
                    self.failed = True
                    raise TrajectoryError(
                        "At least one SEND_TRAJECTORY_COMMAND failed.",
                        self,
                    )

        # Finalise the trajectories
        self.end_traj_cmds = await self.fps.send_command(
            "TRAJECTORY_DATA_END",
            positioner_ids=list(self.trajectories.keys()),
        )

        for cmd in self.end_traj_cmds:
            if cmd.status.failed:
                self.failed = True
                raise TrajectoryError("TRAJECTORY_DATA_END failed.", self)

            if ResponseCode.INVALID_TRAJECTORY in cmd.replies[0].response_code:
                self.failed = True
                raise TrajectoryError(
                    f"positioner_id={cmd.positioner_id} got an "
                    f"INVALID_TRAJECTORY reply.",
                    self,
                )

        self.data_send_time = time.time() - start_trajectory_send_time

        self._ready_to_start = True
        self.failed = False

        return True

    async def start(self, use_sync_line: bool = True):
        """Starts the trajectory."""

        if not self._ready_to_start or self.failed:
            raise TrajectoryError("The trajectory has not been sent.", self)

        if self.move_time is None:
            raise TrajectoryError("move_time not set.", self)

        self.use_sync_line = use_sync_line

        if use_sync_line:

            if not isinstance(self.fps.ieb, IEB) or self.fps.ieb.disabled:
                raise TrajectoryError(
                    "IEB is not connected. Cannot use SYNC line.",
                    self,
                )
            if (await self.fps.ieb.get_device("sync").read())[0] == "closed":
                raise TrajectoryError("The SYNC line is on high.", self)

            sync = self.fps.ieb.get_device("sync")
            assert isinstance(sync, drift.Relay)

            # Set SYNC line to high.
            await sync.close()

            # Schedule reseting of SYNC line
            asyncio.get_event_loop().call_later(0.5, asyncio.create_task, sync.open())

        else:

            # Start trajectories
            command = await self.fps.send_command(
                "START_TRAJECTORY",
                positioner_ids=0,
                timeout=1,
                # All positioners reply, including those not in the trajectory.
                n_positioners=len(self.fps.positioners),
            )

            if command.status.failed:
                await self.fps.stop_trajectory()
                self.failed = True
                raise TrajectoryError("START_TRAJECTORY failed", self)

        restart_pollers = True if self.fps.pollers.running else False
        await self.fps.pollers.stop()

        self.start_time = time.time()

        # min_trajectory_time = 5.0
        # PS = PositionerStatus

        try:

            # # The positioners take a bit to report that they are moving so if the
            # # move time is too short, we don't try to check if the positioners started
            # # moving (but we'll check later that they arrived to their positions.)
            # if self.move_time >= min_trajectory_time:
            #     await asyncio.sleep(min_trajectory_time)
            #     await self.fps.update_status()

            #     if self.fps.status & FPSStatus.IDLE:
            #         raise TrajectoryError("Move failed to start.")

            #     sbits = numpy.array([p.status for p in self.fps.values()])
            #     not_moving = numpy.where(sbits & PS.DISPLACEMENT_COMPLETED)
            #     if not_moving[0].any():
            #         not_moving_pids = numpy.array(list(self.fps))[not_moving[0]]

            #         # Before reporting that the positioner is not moving, check if
            #         # it's already there.
            #         await self.fps.update_position()
            #         positions = self.fps.get_positions()

            #         really_not_moving = []
            #         for pid in not_moving_pids:
            #             if pid not in self.trajectories:
            #                 continue

            #             current = positions[positions[:, 0] == pid][0, 1:]
            #             alpha = self.trajectories[pid]["alpha"][-1][0]
            #             beta = self.trajectories[pid]["beta"][-1][0]

            #             if not numpy.allclose(current - [alpha, beta], 0, atol=0.1):
            #                 really_not_moving.append(pid)

            #         if len(really_not_moving) > 0:
            #             not_moving_str = ", ".join(map(str, really_not_moving))
            #             # Should this be an error?
            #             warnings.warn(
            #                 "Some positioners do not appear to be "
            #                 f"moving: {not_moving_str}.",
            #                 JaegerUserWarning,
            #             )

            while True:

                await asyncio.sleep(1)

                if self.fps.locked:
                    raise TrajectoryError(
                        "The FPS got locked during the trajectory.",
                        self,
                    )

                await self.fps.update_status()

                if self.fps.status & FPSStatus.IDLE:
                    self.failed = False
                    break

                elapsed = time.time() - self.start_time
                if elapsed > (self.move_time + 3):
                    raise TrajectoryError(
                        "Some positioners did not complete the move.",
                        self,
                    )

            # TODO: There seems to be bug in the firmware. Sometimes when a positioner
            # fails to start its trajectory, at the end of the trajectory time it
            # does believe it has reached the commanded position, although it's still
            # at the initial position. In those cases issuing a SEND_TRAJECTORY_ABORT
            # followed by a position update seems to return correct positions.
            await self.fps.stop_trajectory()

            # The FPS says they have all stopped moving but check that they are
            # actually at their positions.
            await self.fps.update_position()
            failed_reach = False
            for pid in self.trajectories:
                alpha = self.trajectories[pid]["alpha"][-1][0]
                beta = self.trajectories[pid]["beta"][-1][0]
                current_position = numpy.array(self.fps[pid].position)
                if not numpy.allclose(current_position, [alpha, beta], atol=0.1):
                    warnings.warn(
                        f"Positioner {pid} may not have reached its destination.",
                        JaegerUserWarning,
                    )
                    failed_reach = True

            if failed_reach:
                raise TrajectoryError(
                    "Some positioners did not reach their destinations.",
                    self,
                )

        except BaseException:
            self.failed = True
            await self.fps.stop_trajectory()
            raise

        finally:
            # Not explicitely updating the positions here because save_snapshot()
            # will do that and no need to waste extra time.
            await self.fps.save_snapshot()

            if self.dump_file:
                self.dump_trajectory()

            self.end_time = time.time()
            if restart_pollers:
                self.fps.pollers.start()

        return True

    def dump_trajectory(self, path: str | None = None):
        """Dumps the trajectory to a JSON file."""

        if path is None:
            if self.dump_file is None:
                return

            path = self.dump_file

        self.dump_data["success"] = not self.failed
        self.dump_data["trajectory_start_time"] = self.start_time
        self.dump_data["trajectory_send_time"] = self.data_send_time
        self.dump_data["use_sync_line"] = self.use_sync_line
        self.dump_data["end_time"] = time.time()
        self.dump_data["final_positions"] = self.fps.get_positions_dict()

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        with open(path, "w") as f:
            f.write(json.dumps(self.dump_data, indent=2))

    async def abort_trajectory(self):
        """Aborts the trajectory transmission."""

        cmd = await self.fps.send_command(
            "SEND_TRAJECTORY_ABORT",
            positioner_ids=list(self.trajectories.keys()),
        )

        if cmd.status.failed:
            raise TrajectoryError("Cannot abort trajectory transmission.", self)


class SendNewTrajectory(Command):
    """Starts a new trajectory and sends the number of points."""

    command_id = CommandID.SEND_NEW_TRAJECTORY
    broadcastable = False
    move_command = True

    def __init__(self, positioner_ids, n_alpha=None, n_beta=None, **kwargs):

        if n_alpha is not None and n_beta is not None:
            kwargs["data"] = self.get_data(n_alpha, n_beta)
        else:
            assert "data" in kwargs, "n_alpha/n_beta or data need to be passed."

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def get_data(n_alpha, n_beta):
        """Returns the data bytearray from a pair for ``(n_alpha, n_beta)``."""

        alpha_positions = int(n_alpha)
        beta_positions = int(n_beta)

        assert alpha_positions > 0 and beta_positions > 0

        data = int_to_bytes(alpha_positions) + int_to_bytes(beta_positions)

        return data


class SendTrajectoryData(Command):
    """Sends a data point in the trajectory.

    This command sends multiple messages that represent a full trajectory.

    Parameters
    ----------
    positions : list
        A list of tuples in which the first element is the angle, in degrees,
        and the second is the associated time, in seconds.

    Examples
    --------
        >>> SendTrajectoryData([(90, 10), (88, 12), (80, 20)])

    """

    command_id = CommandID.SEND_TRAJECTORY_DATA
    broadcastable = False
    move_command = True

    def __init__(self, positioner_ids, positions=None, **kwargs):

        if positions is not None:
            kwargs["data"] = self.calculate_positions(positions)
        else:
            assert "data" in kwargs, "positions or data need to be passed."

        super().__init__(positioner_ids, **kwargs)

    @staticmethod
    def calculate_positions(positions):
        """Converts angle-time posions to bytes data."""

        positions = numpy.array(positions).astype(numpy.float64)

        positions[:, 0] = positions[:, 0] / 360.0 * MOTOR_STEPS
        positions[:, 1] /= TIME_STEP

        positions = positions.astype(numpy.int32)

        data = []
        for angle, tt in positions:
            data.append(int_to_bytes(angle, dtype="i4") + int_to_bytes(tt, dtype="i4"))

        return data


class TrajectoryDataEnd(Command):
    """Indicates that the transmission for the trajectory has ended."""

    command_id = CommandID.TRAJECTORY_DATA_END
    broadcastable = False
    move_command = True


class SendTrajectoryAbort(Command):
    """Aborts sending a trajectory."""

    command_id = CommandID.SEND_TRAJECTORY_ABORT
    broadcastable = False
    move_command = False
    safe = True


class StartTrajectory(Command):
    """Starts the trajectories."""

    command_id = CommandID.START_TRAJECTORY
    broadcastable = True
    move_command = True
    safe = True


class StopTrajectory(Command):
    """Stop the trajectories."""

    command_id = CommandID.STOP_TRAJECTORY
    broadcastable = True
    safe = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-18
# @Filename: configuration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from time import time

from typing import TYPE_CHECKING

import click
import numpy
import peewee
from astropy.time import Time

from sdssdb.peewee.sdss5db import opsdb

from jaeger import config
from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.ieb import IEB
from jaeger.kaiju import check_trajectory
from jaeger.kaiju import explode as kaiju_explode
from jaeger.target.configuration import (
    Configuration,
    DitheredConfiguration,
    ManualConfiguration,
)
from jaeger.target.design import Design
from jaeger.target.tools import copy_summary_file, create_random_configuration
from jaeger.utils.database import get_designid_from_queue, match_assignment_hash

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from jaeger.actor import JaegerActor, JaegerCommandType
    from jaeger.fps import FPS


__all__ = ["configuration"]


async def _load_design(
    command: JaegerCommandType,
    fps: FPS,
    design_id: int | None = None,
    preload: bool = False,
    no_clone: bool = False,
    scale: float | None = None,
    fudge_factor: float | None = None,
    epoch: float | None = None,
    epoch_delay: float = 0.0,
    get_paths: bool = True,
    path_generation_mode: str | None = None,
):
    """Helper to load or preload a design."""

    if design_id is None:
        design_id, _epoch_delay = get_designid_from_queue(
            pop=not preload,
            epoch_delay=True,
        )
        if epoch_delay == 0.0 and _epoch_delay is not None:
            epoch_delay = _epoch_delay

    if design_id is None:
        command.error("Failed getting a new design from the queue.")
        return False

    if preload:
        command.info(f"Preloading design {design_id}.")
    else:
        command.info(f"Loading design {design_id}.")

    valid = Design.check_design(design_id, command.actor.observatory)
    if valid is False:
        command.error(
            "The design does not exists or is not a valid "
            f"{command.actor.observatory} design."
        )
        return False

    # If the original configuration was create longer than max_cloned_time,
    # then do not clone again since we want a chance to update the positions
    # of the robots for the current epoch.
    max_cloned_time = config["configuration"]["max_cloned_time"]

    if (
        no_clone is False
        and fps.configuration is not None
        and time() - fps.configuration.created_time < max_cloned_time
        and fps.configuration.configuration_id is not None
        and fps.configuration.design is not None
        and match_assignment_hash(fps.configuration.design.design_id, design_id)
    ):
        command.info(
            f"Design {design_id} matches previously loaded design "
            f"{fps.configuration.design.design_id}. Cloning configuration."
        )

        configuration = await fps.configuration.clone(
            design_id=design_id,
            write_summary=False,
            write_to_database=False,
        )

    else:

        default_scale = config["configuration"].get("default_focal_scale", 1.0)
        use_guider_scale = config["configuration"].get("use_guider_scale", True)

        clip_scale: float = config["configuration"]["clip_scale"]
        SCALE_FUDGE: float = config["configuration"]["scale_fudge_factor"]
        fudge_factor = fudge_factor or SCALE_FUDGE

        guider_scale: float | None = None

        # Query the guider for the historical scale from the previous exposure.
        if scale is None and use_guider_scale is True:
            command.debug("Getting guider scale.")

            max_scale_age = config["configuration"]["guider_max_scale_age"]
            get_scale_cmd = await command.send_command(
                "cherno",
                f"get-scale --max-age {max_scale_age}",
            )
            if get_scale_cmd.status.did_fail:
                command.warning("Failed getting scale from guider.")
            else:
                guider_scale = float(get_scale_cmd.replies.get("scale_median")[0])
                if guider_scale < 0:
                    command.warning("Invalid guider scale.")
                    guider_scale = None
                else:
                    # Clip scale if needed.
                    if (abs(guider_scale) - 1) * 1e6 > clip_scale:
                        guider_scale = numpy.clip(
                            guider_scale,
                            1 - clip_scale / 1e6,
                            1 + clip_scale / 1e6,
                        )
                        command.warning(
                            "Unexpectedly large guider scale. "
                            f"Clipping to {guider_scale:.6f}."
                        )

            temp_coeffs = config["configuration"].get("scale_temperature_coeffs", None)

            if guider_scale is None and temp_coeffs is not None:
                # Try using the scale-temperature relationship instead.
                try:
                    if not isinstance(command.actor.fps.ieb, IEB):
                        raise ValueError("IEB not connected")

                    temperature = (await command.actor.fps.ieb.read_device("T3"))[0]
                    if not isinstance(temperature, float) or temperature < -100:
                        raise ValueError("Invalid ambient temperature")

                    guider_scale = float(numpy.polyval(temp_coeffs, temperature))

                    command.debug(
                        "Using focal scale factor derived from ambient "
                        f"temperature ({temperature:.2f} C): {guider_scale:.6f}"
                    )
                except Exception as err:
                    command.warning(
                        f"Failed getting ambient temperature: {err} "
                        "Using default scale."
                    )

        # Second pass at setting the scale. Apply the fudge factor.
        if guider_scale is not None:
            scale = guider_scale * fudge_factor
            command.debug(
                f"Correcting guider scale {guider_scale:.6f} with fudge factor "
                f"{fudge_factor:.6f}."
            )

        if scale is not None:
            command.info(f"Focal scale: {scale:.6f}.")
        else:
            scale = default_scale
            command.info(f"Using default scale {scale:.6f}.")

        # Define the epoch for the configuration.
        try:
            command.debug(text=f"Epoch delay: {round(epoch_delay, 1)} seconds.")

            if epoch is None:
                epoch = float(Time.now().jd)
            epoch += epoch_delay / 86400.0

            design = await Design.create_async(design_id, epoch=epoch, scale=scale)
        except Exception as err:
            command.error(error=f"Failed retrieving design: {err}")
            return False

        configuration = design.configuration

    if get_paths and configuration.is_cloned is False:
        command.info("Calculating trajectories.")
        configuration.set_command(command)
        await configuration.get_paths(
            decollide=True,
            path_generation_mode=path_generation_mode,
        )

    return configuration


@jaeger_parser.group()
def configuration():
    """Allows to load a configuration, ingest it to the database, and execute it."""
    pass


@configuration.command(cancellable=True)
@click.option(
    "--reload",
    is_flag=True,
    help="If the design is currently loaded, creates a new configuration.",
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace an existing entry.",
)
@click.option(
    "--generate-paths/--no-generate-paths",
    default=True,
    help="Generates and stores the to and from destination paths.",
)
@click.option(
    "--epoch",
    type=float,
    help="Epoch at which to create the configuration, as a Julian Day.",
)
@click.option(
    "--epoch-delay",
    type=float,
    default=0.0,
    help="A delay in seconds for the epoch for which the configuration is calculated.",
)
@click.option(
    "--from-positions",
    is_flag=True,
    help="Loads a configuration from the current robot positions.",
)
@click.option(
    "--from-preloaded",
    is_flag=True,
    help="Finishes loading a preloaded configuration.",
)
@click.option(
    "--ingest/--no-ingest",
    default=True,
    help="Whether to ingest the configuration into the DB.",
)
@click.option(
    "--write-summary/--no-write-summary",
    default=True,
    help="Whether to write the summary file for the configuration. "
    "Ignored if --no-ingest.",
)
@click.option(
    "--execute/--no-execute",
    default=False,
    help="Send and start the from_destination trajectory.",
)
@click.option(
    "--reissue",
    is_flag=True,
    help="Only reissue the configuration_loaded keyword.",
)
@click.option(
    "--scale",
    type=float,
    help="Focal plane scale factor. If not passes, uses coordio default.",
)
@click.option(
    "--fudge-factor",
    type=float,
    default=None,
    help="Modify the default scale fudge factor.",
)
@click.option(
    "--no-clone",
    is_flag=True,
    help="If the new design has the same target set as the currently loaded one, "
    "does not clone the configuration and instead loads the new design.",
)
@click.option(
    "--path-generation-mode",
    type=click.Choice(["greedy", "mdp"], case_sensitive=False),
    help="The path generation algorithm to use.",
)
@click.argument("DESIGNID", type=int, required=False)
async def load(
    command: Command[JaegerActor],
    fps: FPS,
    designid: int | None = None,
    reload: bool = False,
    replace: bool = False,
    from_positions: bool = False,
    from_preloaded: bool = False,
    generate_paths: bool = False,
    epoch: float | None = None,
    epoch_delay: float = 0.0,
    ingest: bool = False,
    write_summary: bool = False,
    execute: bool = False,
    reissue: bool = False,
    scale: float | None = None,
    fudge_factor: float | None = None,
    no_clone: bool = False,
    path_generation_mode: str | None = None,
):
    """Creates and ingests a configuration from a design in the database."""

    assert command.actor is not None

    if reissue is True:
        if fps.configuration is None or fps.configuration.design is None:
            return command.fail("No configuration loaded.")
        if designid is not None and designid != fps.configuration.design.design_id:
            return command.fail("Mismatch between loaded and provided design IDs.")
        _output_configuration_loaded(command, fps)
        return command.finish()

    if reload is True:
        if fps.configuration is None:
            return command.fail(error="No configuration found. Cannot reload.")
        designid = fps.configuration.design_id
        fps.configuration.configuration_id = None

    elif from_preloaded is True:
        if fps._preloaded_configuration is None:
            return command.fail("No preloaded configuration available.")

        fps.configuration = fps._preloaded_configuration

    elif from_positions is True:
        await fps.update_position()
        positions = fps.get_positions_dict()
        fps.configuration = ManualConfiguration.create_from_positions(positions)

    else:
        configuration = await _load_design(
            command,
            fps,
            design_id=designid,
            fudge_factor=fudge_factor,
            preload=False,
            no_clone=no_clone,
            scale=scale,
            epoch=epoch,
            epoch_delay=epoch_delay,
            get_paths=False,
            path_generation_mode=path_generation_mode,
        )

        if configuration is False:
            # _load_design already issues an error so we just fail.
            return command.fail()
        else:
            fps.configuration = configuration

    assert isinstance(fps.configuration, (Configuration, ManualConfiguration))

    if fps.configuration is None:
        return command.fail(error="A configuration must first be loaded.")

    if fps.configuration.ingested is False:
        replace = False

    fps.configuration.set_command(command)

    if (
        not fps.configuration.is_cloned
        and generate_paths
        and fps.configuration.to_destination is None
        and not from_positions
    ):
        try:
            command.info("Calculating trajectories.")
            await fps.configuration.get_paths(
                decollide=not from_positions,
                path_generation_mode=path_generation_mode,
            )
        except Exception as err:
            return command.fail(error=f"Failed generating paths: {err}")

    if ingest:
        fps.configuration.write_to_database(replace=replace)
    else:
        command.warning("Not ingesting configuration. Configuration ID is -999.")
        fps.configuration.configuration_id = -999

    if ingest and write_summary:
        await fps.configuration.write_summary(
            overwrite=True,
            headers={"cloned_from": fps.configuration.cloned_from or -999},
        )

        if fps.configuration.cloned_from and fps.configuration.configuration_id:
            copy_summary_file(
                fps.configuration.cloned_from,
                fps.configuration.configuration_id,
                fps.configuration.design_id,
                "F",
            )

    _output_configuration_loaded(command, fps)

    snapshot = await fps.configuration.save_snapshot()
    command.info(configuration_snapshot=snapshot)

    if execute and not from_positions:
        cmd = await command.send_command("jaeger", "configuration execute")
        if cmd.status.did_fail:
            if cmd.status.is_done:
                return
            else:
                return cmd.fail("Failed executing configuration.")

    fps._preloaded_configuration = None
    command.debug(design_preloaded=-999)

    if fps.configuration.design_id is not None and fps.configuration.design_id > 0:
        design_id = fps.configuration.design_id
        # Check if the design is in the queue and it would be the next one to pop.
        # If so, pop it.
        try:
            queue_instance = (
                opsdb.Queue.select()
                .where(opsdb.Queue.design_id == design_id)
                .order_by(opsdb.Queue.position)
                .first()
            )
            if queue_instance and queue_instance.position == 1:
                opsdb.Queue.pop()
        except peewee.DoesNotExist:
            pass

    return command.finish(f"Configuration {fps.configuration.configuration_id} loaded.")


@configuration.command()
async def reload(command: Command[JaegerActor], fps: FPS):
    """Reloads the current configuration.

    This is equivalent to using jaeger configuration load --no-clone DESIGNID
    where DESIGNID is the currently loaded design.

    """

    if (
        fps.configuration is None
        or fps.configuration.design_id is None
        or fps.configuration.design_id == -999
    ):
        return command.fail("There is no configuration currently loaded.")

    design_id = fps.configuration.design_id
    await command.child_command(f"configuration load --no-clone {design_id}")

    return command.finish()


def _output_configuration_loaded(command: Command[JaegerActor], fps: FPS):
    """Outputs the loaded configuration."""

    assert fps.configuration

    command.debug(text=f"Focal plane scale: {fps.configuration.assignment_data.scale}.")

    boresight = fps.configuration.assignment_data.boresight
    configuration = fps.configuration

    command.debug(
        configuration_loaded=[
            configuration.configuration_id,
            configuration.design.design_id if configuration.design else -999,
            configuration.design.field.field_id if configuration.design else -999,
            configuration.design.field.racen if configuration.design else -999.0,
            configuration.design.field.deccen if configuration.design else -999.0,
            configuration.design.field.position_angle if configuration.design else 0,
            boresight[0, 0] if boresight is not None else -999.0,
            boresight[0, 1] if boresight is not None else -999.0,
            configuration._summary_file or "",
            configuration.is_cloned,
        ]
    )


@configuration.command()
async def clone(command: Command[JaegerActor], fps: FPS):
    """Clones a configuration."""

    if fps.configuration is None:
        return command.fail(error="A configuration must first be loaded.")

    new = await fps.configuration.clone()
    fps.configuration = new

    _output_configuration_loaded(command, fps)

    return command.finish()


@configuration.command(cancellable=True)
@click.option(
    "--epoch",
    type=float,
    help="Epoch at which to create the configuration, as a Julian Day.",
)
@click.option(
    "--epoch-delay",
    type=float,
    default=0.0,
    help="A delay in seconds for the epoch for which the configuration is calculated.",
)
@click.option(
    "--scale",
    type=float,
    help="Focal plane scale factor. If not passes, uses coordio default.",
)
@click.option(
    "--fudge-factor",
    type=float,
    default=None,
    help="Modify the default scale fudge factor.",
)
@click.option(
    "--no-clone",
    is_flag=True,
    help="If the new design has the same target set as the currently loaded one, "
    "does not clone the configuration and instead loads the new design.",
)
@click.option(
    "--make-active",
    is_flag=True,
    help="Loads the configuration after preloading.",
)
@click.option(
    "--clear",
    is_flag=True,
    help="Clears the preloaded configuration.",
)
@click.argument("DESIGNID", type=int, required=False)
async def preload(
    command: JaegerCommandType,
    fps: FPS,
    designid: int | None = None,
    epoch: float | None = None,
    epoch_delay: float = 0.0,
    scale: float | None = None,
    fudge_factor: float | None = None,
    no_clone: bool = False,
    make_active: bool = True,
    clear: bool = False,
):
    """Preloads a design.

    Preloading a design works similar to loading it, but no files are generated, no
    database entry is created, no configuration_loaded keyword is output, and the
    new configuration is stored separately. To make the preloaded configuration
    active either pass the --make-active flag or run
    jaeger configuration load --from-preloaded.

    """

    if clear:
        fps._preloaded_configuration = None
        return command.finish(design_preloaded=-999)

    configuration = await _load_design(
        command,
        fps,
        design_id=designid,
        preload=True,
        no_clone=no_clone,
        scale=scale,
        fudge_factor=fudge_factor,
        epoch=epoch,
        epoch_delay=epoch_delay,
    )

    if configuration is False:
        # _load_design already issues an error so we just fail.
        return command.fail()
    elif configuration.design_id is None:
        return command.fail(
            "Preloaded configuration does not have design ID. "
            "This should never happen."
        )
    else:
        fps._preloaded_configuration = configuration

    command.info(design_preloaded=configuration.design_id)

    if make_active:
        load_cmd = await command.send_command(
            "jaeger",
            "configuration load --from-preloaded",
        )
        if load_cmd.status.did_fail:
            return command.fail("Failed making the preloaded configuration active.")

    return command.finish()


@configuration.command()
async def execute(command: Command[JaegerActor], fps: FPS):
    """Executes a configuration trajectory (folded to targets)."""

    if fps.locked:
        command.fail(error="The FPS is locked.")

    if fps.configuration is None:
        return command.fail(error="A configuration must first be loaded.")

    if fps.configuration.is_dither:
        assert isinstance(fps.configuration, DitheredConfiguration)
        if fps.configuration.to_destination is not None:
            trajectory = fps.configuration.to_destination
            command.info(text="Using stored trajectory (to destination).")
        else:
            command.info(text="Calculating trajectory.")
            trajectory = await fps.configuration.get_paths()
    else:
        if fps.configuration.from_destination is not None:
            trajectory = fps.configuration.from_destination
            command.info(text="Using stored trajectory (from destination).")
        else:
            command.info(text="Calculating trajectory.")

            try:
                trajectory = await fps.configuration.get_paths()
            except Exception as err:
                return command.fail(error=f"Failed getting trajectory: {err}")

        if not (await check_trajectory(trajectory, fps=fps, atol=1)):
            return command.fail(error="Trajectory validation failed.")

    command.info(text="Sending and executing forward trajectory.")

    try:
        await fps.send_trajectory(trajectory, command=command)
    except TrajectoryError as err:
        return command.fail(error=f"Trajectory failed with error: {err}")

    fps.configuration.executed = True

    command.finish()


@configuration.command()
@click.option("--explode", is_flag=True, help="Explode before reversing.")
async def reverse(command: Command[JaegerActor], fps: FPS, explode: bool = False):
    """Executes a reverse trajectory (targets to folded)."""

    if fps.locked:
        return command.fail(error="The FPS is locked.")

    if await fps.is_folded():
        return command.finish("Already folded.")

    configuration = fps.configuration
    if (fps.configuration is None or fps.configuration.executed is False) and (
        len(fps._previous_configurations) > 0
        and fps._previous_configurations[-1] is not None
        and fps._previous_configurations[-1].executed is True
    ):
        configuration = fps._previous_configurations[-1]
        command.warning("Using previous configuration reverse path.")

    if configuration is None:
        return command.fail(error="A configuration must first be loaded.")

    if configuration.executed is False:
        return command.fail(error="The configuration was not executed. Try unwinding.")

    await fps.update_position()

    if configuration.is_dither:
        trajectory = configuration.from_destination
    else:
        trajectory = configuration.to_destination
        if trajectory is None:
            return command.fail(
                error="The configuration does not have a to_destination "
                "path. Use unwind."
            )

        try:
            if explode:
                # First we explode the robots a bit.
                command.info("Exploding before reversing.")
                current_positions = fps.get_positions_dict()
                explode_path = await kaiju_explode(
                    current_positions,
                    5.0,
                    disabled=[pid for pid in fps.positioners if fps[pid].disabled],
                )
                await fps.send_trajectory(explode_path, command=command)

                # Then we send a goto to the initial point of the reverse trajectory.
                new_positions = {
                    pid: (trajectory[pid]["alpha"][0][0], trajectory[pid]["beta"][0][0])
                    for pid in trajectory
                }
                command.info("Reverting to final configuration positions.")
                await fps.goto(new_positions)

        except Exception as err:
            return command.fail(f"Failed preparing to send reverse trajectory: {err}")

    command.info(text="Sending and executing reverse trajectory.")

    try:
        await fps.send_trajectory(trajectory, command=command)
    except TrajectoryError as err:
        return command.fail(error=f"Trajectory failed with error: {err}")

    if configuration.is_dither:
        command.info("Restoring parent configuration.")
        fps.configuration = configuration.parent_configuration
        _output_configuration_loaded(command, fps)

    command.finish()


@configuration.command()
@click.argument("RADIUS", type=float)
@click.option(
    "--epoch",
    type=str,
    default="inherit",
    help="Epoch to use for the dithered configuration. With 'inherit', the parent "
    "configuration epoch is used; 'now' uses the current time. Otherwise provide a JD.",
)
async def dither(
    command: Command[JaegerActor],
    fps: FPS,
    radius: float,
    epoch: str = "inherit",
):
    """Dither a loaded configuration."""

    if fps.configuration is None:
        return command.fail("A configuration must first be loaded.")

    is_dithered = isinstance(fps.configuration, DitheredConfiguration)
    if is_dithered and fps.configuration.parent_configuration is not None:
        parent_configuration = fps.configuration.parent_configuration
    else:
        parent_configuration = fps.configuration

    configuration_id = parent_configuration.configuration_id
    command.info(f"Creating dithered configuration from {configuration_id}")

    epoch = epoch.lower()
    if epoch == "inherit":
        dither_epoch = None
    elif epoch == "now":
        dither_epoch = Time.now().jd
    else:
        dither_epoch = float(epoch)

    fps.configuration = DitheredConfiguration(
        parent_configuration,
        radius,
        epoch=dither_epoch,
    )
    await fps.configuration.get_paths()

    fps.configuration.write_to_database()
    await fps.configuration.write_summary(overwrite=True)

    _output_configuration_loaded(command, fps)

    command.info("Executing dithered configuration.")
    execute_cmd = await command.send_command("jaeger", "configuration execute")
    if execute_cmd.status.did_fail:
        command.fail("Failed executing configuration.")

    fps.configuration.executed = True

    configuration_id = fps.configuration.configuration_id
    command.finish(f"Dithered configuration {configuration_id} loaded and executed.")


@configuration.command()
@click.argument("POSITION_ANGLE", type=float, required=False)
async def slew(
    command: Command[JaegerActor],
    fps: FPS,
    position_angle: float | None = None,
):
    """Slews to the field centre of a configuration.

    Optionally pass the position angle of the rotator in DEGREES.

    """

    command.warning(
        "jaeger configuration slew is a temporary command that does not "
        "perform any safety checks. It will be replaced with hal goto."
    )

    if fps.configuration is None:
        return command.fail("Configuration not loaded.")

    if isinstance(fps.configuration, DitheredConfiguration):
        return command.fail("Cannot slew to a dithered configuration.")

    if fps.configuration.design is None:
        return command.fail("The configuration does not have a design.")

    ra = fps.configuration.design.field.racen
    dec = fps.configuration.design.field.deccen

    if position_angle is not None:
        pa = position_angle
    else:
        pa = fps.configuration.design.field.position_angle

    command.info(f"Slewing to ({ra}, {dec}, {pa})")

    if fps.observatory == "APO":
        slew_cmd = await command.send_command(
            "tcc",
            f"track {ra}, {dec} icrs /rottype=object /rotang={pa:g} /rotwrap=mid",
        )
    else:
        slew_cmd = await command.send_command(
            "lcotcc",
            f"target {ra}, {dec} /posAngle={pa:.3f}",
        )

    if slew_cmd.status.did_fail:
        return command.fail("Failed slewing to field.")

    return command.finish("Slew complete.")


@configuration.command()
@click.argument("SEED", type=int, required=False)
@click.option("--danger", is_flag=True, help="Use full range of alpha and beta.")
@click.option(
    "--uniform",
    type=str,
    default=None,
    help="Comma-separated alpha and beta ranges.",
)
@click.option(
    "--collision-buffer",
    type=click.FloatRange(1.6, 3.0),
    help="Custom collision buffer",
)
@click.option(
    "--send-trajectory/--no-send-trajectory",
    help="Send the trajectory to the FPS.",
    default=True,
)
@click.option(
    "--max-retries",
    type=int,
    default=10,
    help="Home many retries to allow due to deadlocks.",
)
@click.option(
    "--path-generation-mode",
    type=click.Choice(["greedy", "mdp"], case_sensitive=False),
    help="The path generation algorithm to use.",
)
async def random(
    command: Command[JaegerActor],
    fps: FPS,
    seed: int | None = None,
    danger: bool = False,
    uniform: str | None = None,
    collision_buffer: float | None = None,
    send_trajectory: bool = True,
    max_retries: int = 10,
    path_generation_mode: str | None = None,
):
    """Executes a random, valid configuration."""

    if fps.locked:
        command.fail(error="The FPS is locked.")

    command.debug(text="Checking that all positioners are folded.")

    # Check that all positioners are folded.
    await fps.update_position()
    positions = fps.get_positions(ignore_disabled=True)

    if len(positions) == 0:
        return command.fail("No positioners connected")

    alphaL: float
    betaL: float

    alphaL, betaL = config["kaiju"]["lattice_position"]
    if not numpy.allclose(positions[:, 1:] - [alphaL, betaL], 0, atol=1):
        return command.fail(error="Not all the positioners are folded.")

    command.info(text="Creating random configuration.")

    uniform_unpack: tuple[float, ...] | None
    if uniform:
        try:
            uniform_unpack = tuple(map(float, uniform.split(",")))
        except Exception as err:
            return command.fail(error=f"Failed calculating uniform ranges: {err}")
    else:
        uniform_unpack = None

    try:
        configuration = await create_random_configuration(
            fps,
            seed=seed,
            uniform=uniform_unpack,
            safe=not danger,
            collision_buffer=collision_buffer,
            max_retries=max_retries,
        )

        command.info("Getting trajectory.")
        trajectory = await configuration.get_paths(
            decollide=False,
            collision_buffer=collision_buffer,
            path_generation_mode=path_generation_mode,
        )
    except JaegerError as err:
        return command.fail(error=f"jaeger random failed: {err}")

    # Make this the FPS configuration
    assert command.actor
    command.actor.fps.configuration = configuration

    if send_trajectory is False:
        return command.finish()

    command.info("Executing random trajectory.")

    try:
        await fps.send_trajectory(trajectory, command=command)
    except TrajectoryError as err:
        return command.fail(error=f"Trajectory failed with error: {err}")

    configuration.executed = True

    command.finish()


@configuration.command()
@click.argument("RA", type=float)
@click.argument("DEC", type=float)
@click.argument("PA", type=float)
async def fake_field(
    command: Command[JaegerActor],
    fps: FPS,
    ra: float,
    dec: float,
    pa: float,
):
    """Output a fake field."""

    command.info(
        configuration_loaded=[
            -999,
            -999,
            -999,
            ra,
            dec,
            pa,
            -999.0,
            -999.0,
            "?",
            False,
        ]
    )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-11
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os

from typing import Optional, Tuple

import numpy
from astropy.time import Time

from jaeger import config
from jaeger.exceptions import JaegerError
from jaeger.maskbits import ResponseCode


__all__ = [
    "get_dtype_str",
    "int_to_bytes",
    "bytes_to_int",
    "get_identifier",
    "parse_identifier",
    "motor_steps_to_angle",
    "get_goto_move_time",
    "get_sjd",
]


MOTOR_STEPS = config["positioner"]["motor_steps"]


def get_dtype_str(dtype, byteorder="little"):
    """Parses dtype and byte order to return a type string code.

    Parameters
    ----------
    dtype : `numpy.dtype` or `str`
        Either a dtype (e.g., ``numpy.uint32``) or a string with the type code
        (``'>u4'``). If a string type code and the first character indicates
        the byte order (``'>'`` for big, ``'<'`` for little endian),
        ``byteorder`` will be ignored. The type code refers to bytes, while the
        dtype classes refer to bits, i.e., ``'u2'`` is equivalent to
    byteorder : str
        Either ``'big'`` for big endian representation or ``'little'`` for
        little end. ``'>'`` and ``'<'`` are also accepted, respectively.

    Returns
    -------
    type_code : `str`
        The type code for the input dtype and byte order.

    Examples
    --------
    ::

        >>> get_dtype_str(numpy.uint32, byteorder='big')
        '>u4'
        >>> get_dtype_str('u2', byteorder='>')
        '>u2'
        >>> get_dtype_str('<u2', byteorder='big')
        '<u2'

    """

    if byteorder == "big":
        byteorder = ">"
    elif byteorder == "little":
        byteorder = "<"
    elif byteorder in [">", "<"]:
        pass
    else:
        raise ValueError(f"invalid byteorder {byteorder}")

    if isinstance(dtype, str):
        if dtype[0] in [">", "<"]:
            return dtype
        elif dtype[0] == "=":
            raise ValueError("invalid byte order =. Please, use a specific endianess.")
        else:
            return byteorder + dtype

    dtype_str = dtype().dtype.str

    return byteorder + dtype_str[1:]


def int_to_bytes(value, dtype="u4", byteorder="little"):
    r"""Returns a bytearray with the representation of an integer.

    Parameters
    ----------
    value : int
        The integer to convert to bytes.
    dtype : `numpy.dtype` or `str`
        The `numpy.dtype` of the byte representation for the integer, or a
        type code that can include the endianess. See `.get_dtype_str` to
        understand how ``dtype`` and ``byteorder`` will be parsed.
    byteorder : str
        Either ``'big'`` for big endian representation or ``'little'`` for
        little end. ``'>'`` and ``'<'`` are also accepted, respectively.

    Returns
    -------
    bytes : `bytearray`
        A `bytearray` with the representation for the input integer.

    Examples
    --------
    ::

        >>> int_to_bytes(5, dtype=numpy.uint16, byteorder='big')
        bytearray(b'\x00\x05')

    """

    type_code = get_dtype_str(dtype, byteorder=byteorder)

    np_value = numpy.array(value, dtype=type_code)

    return bytearray(np_value.tobytes())


def bytes_to_int(bytes, dtype="u4", byteorder="little"):
    r"""Returns the integer from a bytearray representation.

    Parameters
    ----------
    bytes : `bytearray`
        The bytearray representing the integer.
    dtype : `numpy.dtype` or `str`
        The `numpy.dtype` of the byte representation for the integer, or a
        type code that can include the endianess. See `.get_dtype_str` to
        understand how ``dtype`` and ``byteorder`` will be parsed.
    byteorder : str
        Either ``'big'`` for big endian representation or ``'little'`` for
        little end. ``'>'`` and ``'<'`` are also accepted, respectively.

    Returns
    -------
    integer : int
        A integer represented by ``bytes``.

    Examples
    --------
    ::

        >>> bytes_to_int(b'\x00\x05', dtype=numpy.uint16, byteorder='big')
        5

    """

    type_code = get_dtype_str(dtype, byteorder=byteorder)

    np_buffer = numpy.frombuffer(bytes, dtype=type_code)

    return np_buffer[0]


def get_identifier(positioner_id, command_id, uid=0, response_code=0):
    """Returns a 29 bits identifier with the correct format.

    The CAN identifier format for the positioners uses an extended frame with
    29-bit encoding so that the 11 higher bits correspond to the positioner
    ID, the 8 middle bits are the command number, the following 6 bits are the
    unique identifier, and the 4 lower bits are the response code.

    Parameters
    ----------
    positioner_id : int
        The Id of the positioner to command, or zero for broadcast.
    command_id : int
        The ID of the command to send.
    uid : int
        The unique identifier
    response_code : int
        The response code.

    Returns
    -------
    identifier : `int`
        The decimal integer corresponding to the 29-bit identifier.

    Examples
    --------
    ::

        >>> get_identifier(5, 17, uid=5)
        1328128
        >>> bin(1328128)
        '0b101000100010000000000'

    """

    posid_bin = format(positioner_id, "011b")
    cid_bin = format(command_id, "08b")
    cuid_bin = format(uid, "06b")
    response_bin = format(int(response_code), "04b")

    identifier = posid_bin + cid_bin + cuid_bin + response_bin

    assert len(identifier) == 29

    return int(identifier, 2)


def parse_identifier(identifier: int) -> Tuple[int, int, int, ResponseCode]:
    """Parses an extended frame identifier and returns its components.

    The 29-bit extended frame identifier is composed of a positioner id,
    a command id, and a response code. This function parses an identifier
    and returns the value of each element.

    Parameters
    ----------
    identifier
        The identifier returned by the CAN bus.

    Returns
    -------
    components
        A tuple with the components of the identifier. The first element is
        the positioner id, the second the command id, the third is the command
        UID, and the last one is the response flag as an instance of
        `~jaeger.maskbits.ResponseCode`.

    Examples
    --------
    ::

        >>> parse_identifier(1315072)
        (5, 17, <ResponseCode.COMMAND_ACCEPTED: 0>)
        >>> parse_identifier(1315074)
        (5, 17, <ResponseCode.INVALID_TRAJECTORY: 2>)

    """

    def last(k, n):
        return (k) & ((1 << (n)) - 1)

    def mid(k, m, n):
        return last((k) >> (m), ((n) - (m)))

    positioner_id = mid(identifier, 18, 29)
    command_id = mid(identifier, 10, 18)
    command_uid = mid(identifier, 4, 10)
    response_code = mid(identifier, 0, 4)

    response_flag = ResponseCode(response_code)

    return positioner_id, command_id, command_uid, response_flag


def motor_steps_to_angle(alpha, beta, motor_steps=None, inverse=False):
    """Converts motor steps to angles or vice-versa.

    Parameters
    ----------
    alpha : float
        The alpha position.
    beta : float
        The beta position.
    motor_steps : int
        The number of steps in the motor. Defaults to
        the configuration value ``positioner.moter_steps``.
    inverse : bool
        If `True`, converts from angles to motor steps.

    Returns
    -------
    angles : `tuple`
        A tuple with the alpha and beta angles associated to the input
        motor steps. If ``inverse=True``, ``alpha`` and ``beta`` are considered
        to be angles and the associated motor steps are returned.

    """

    motor_steps = motor_steps or MOTOR_STEPS

    if inverse:
        return (
            int(numpy.round(alpha * motor_steps / 360.0)),
            int(numpy.round(beta * motor_steps / 360.0)),
        )

    return alpha / motor_steps * 360.0, beta / motor_steps * 360.0


def get_goto_move_time(move, speed=None):
    r"""Returns the approximate time need for a given move, in seconds.

    The move time is calculated as :math:`\dfrac{60 \alpha r}{360 v}` where
    :math:`\alpha` is the angle, :math:`r` is the reduction ratio, and
    :math:`v` is the speed in the input in RPM. It adds 0.25s due to
    deceleration; this value is not exact but it's a good approximation for
    most situations.

    Parameters
    ----------
    move : float
        The move, in degrees.
    speed : float
        The speed of the motor for the move, in RPM on the input.

    """

    speed = speed or config["positioner"]["motor_speed"]

    return move * config["positioner"]["reduction_ratio"] / (6.0 * speed) + 0.25


def get_sjd(observatory: Optional[str] = None) -> int:
    """Returns the SDSS Julian Date as an integer based on the observatory.

    Parameters
    ----------
    observatory
        The current observatory, either APO or LCO. If `None`, uses ``$OBSERVATORY``.

    """

    if observatory is None:
        try:
            observatory = os.environ["OBSERVATORY"]
        except KeyError:
            raise JaegerError("Observatory not passed and $OBSERVATORY is not set.")

    observatory = observatory.upper()
    if observatory not in ["APO", "LCO"]:
        raise JaegerError(f"Invalid observatory {observatory}.")

    time = Time.now()
    mjd = time.mjd

    if observatory == "APO":
        return int(mjd + 0.3)
    else:
        return int(mjd + 0.4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-07-27
# @Filename: fvc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, cast

import click

from drift import DriftError, Relay

from jaeger import config
from jaeger.exceptions import FVCError
from jaeger.fvc import FVC
from jaeger.ieb import FVC as FVC_IEB
from jaeger.target.configuration import ManualConfiguration
from jaeger.utils import run_in_executor

from . import jaeger_parser


if TYPE_CHECKING:
    from clu import Command

    from jaeger import FPS
    from jaeger.actor import JaegerActor, JaegerCommandType
    from jaeger.target.configuration import BaseConfiguration


__all__ = ["fvc_parser"]


@jaeger_parser.group(name="fvc")
def fvc_parser():
    """Commands to command the FVC."""

    pass


@fvc_parser.command()
@click.argument("EXPOSURE-TIME", default=None, type=float, required=False)
async def expose(
    command: Command[JaegerActor],
    fps: FPS,
    exposure_time: Optional[float] = None,
):
    """Takes an exposure with the FVC."""

    exposure_time = exposure_time or config["fvc"]["exposure_time"]
    assert isinstance(exposure_time, float)

    command.info("Taking FVC exposure with fliswarm.")

    fvc = FVC(config["observatory"])

    try:
        fvc.set_command(command)
        filename = await fvc.expose(exposure_time=exposure_time)
    except FVCError as err:
        return command.fail(error=f"Failed taking FVC exposure: {err}")

    return command.finish(fvc_filename=str(filename))


@fvc_parser.command(cancellable=True)
@click.option(
    "--exposure-time",
    type=float,
    help="Exposure time.",
)
@click.option(
    "--fbi-level",
    type=float,
    help="FBI LED levels.",
)
@click.option(
    "--one",
    is_flag=True,
    help="Only runs one FVC correction iteration.",
)
@click.option(
    "--max-iterations",
    type=int,
    help="Maximum number of iterations.",
)
@click.option(
    "--stack",
    type=int,
    default=1,
    help="Number of FVC image to stack.",
)
@click.option(
    "--plot/--no-plot",
    default=True,
    help="Generate and save plots.",
)
@click.option(
    "--apply/--no-apply",
    default=True,
    help="Apply corrections.",
)
@click.option(
    "-m",
    "--max-correction",
    type=float,
    help="Maximum correction allowed, in degrees.",
)
@click.option(
    "--target-90-percentile",
    type=float,
    help="90% percentile target, in microns, at which to stop iterating.",
)
@click.option(
    "-k",
    type=float,
    help="Proportional term of the correction.",
)
@click.option(
    "--centroid-method",
    type=click.Choice(["nudge", "simple", "winpos", "sep", "zbplus", "zbminus"]),
    help="The centroid method used to extract sources.",
)
@click.option(
    "--use-invkin/--no-use-invkin",
    default=True,
    help="Use new inverse kinnematics.",
)
@click.option(
    "--no-write-summary",
    is_flag=True,
    help="Does not try to write a confSummaryF file.",
)
async def loop(command: Command[JaegerActor], fps: FPS, **kwargs):
    """Executes the FVC correction loop.

    This routine will turn the FBI LEDs on, take FVC exposures, process them,
    calculate the offset correction and applies them. Loops until the desided
    convergence is achieved.

    """

    result = await take_fvc_loop(
        command,
        fps,
        configuration=fps.configuration,
        **kwargs,
    )

    if result is True:
        return command.finish()
    else:
        return command.fail("The FVC loop failed.")


@fvc_parser.command()
async def status(command: Command[JaegerActor], fps: FPS):
    """Reports the status of the FVC."""

    fvc_ieb = FVC_IEB.create()

    try:
        status = {}
        categories = fvc_ieb.get_categories()
        for category in sorted(categories):
            cat_data = await fvc_ieb.read_category(category)
            status[category] = []
            for cd in cat_data:
                value = cat_data[cd][0]
                if value == "closed":
                    value = True
                elif value == "open":
                    value = False
                else:
                    value = round(value, 1)
                status[category].append(value)

        command.finish(status)

    except DriftError:
        return command.fail(error="FVC IEB is unavailable or failed to connect.")


async def _power_device(device: str, mode: str):
    """Power on/off the device."""

    fvc_ieb = FVC_IEB.create()

    dev: Relay = cast(Relay, fvc_ieb.get_device(device))
    if mode == "on":
        await dev.close()
    else:
        await dev.open()


async def _execute_on_off_command(
    command: Command[JaegerActor], device: str, mode: str
):
    """Executes the on/off command."""

    mode = mode.lower()

    try:
        await _power_device(device, mode)
        command.info(text=f"{device} is now {mode}.")
    except DriftError:
        return command.fail(error=f"Failed to turn {device} {mode}.")

    await command.send_command("jaeger", "fvc status")

    return command.finish()


@fvc_parser.command()
@click.argument("MODE", type=click.Choice(["on", "off"], case_sensitive=False))
async def camera(command: Command[JaegerActor], fps: FPS, mode: str):
    """Turns camera on/off."""

    await _execute_on_off_command(command, "FVC", mode)


@fvc_parser.command()
@click.argument("MODE", type=click.Choice(["on", "off"], case_sensitive=False))
async def NUC(command: Command[JaegerActor], fps: FPS, mode: str):
    """Turns NUC on/off."""

    await _execute_on_off_command(command, "NUC", mode)


@fvc_parser.command()
@click.argument("LEVEL", type=int)
async def led(command: Command[JaegerActor], fps: FPS, level: int):
    """Sets the level of the FVC LED."""

    fvc_ieb = FVC_IEB.create()
    led = fvc_ieb.get_device("LED1")

    raw_value = 32 * int(1023 * (level / 100))
    await led.write(raw_value)

    await command.send_command("jaeger", "fvc status")

    return command.finish()


@fvc_parser.command()
async def snapshot(command: JaegerCommandType, fps: FPS):
    """Takes an FPS snapshot with the FVC. Roughly equivalent to fvc loop --no-apply."""

    # Create a configuration from positions but don't make it active in the FPS.
    positions = fps.get_positions_dict()
    configuration = ManualConfiguration.create_from_positions(positions)

    result = await take_fvc_loop(
        command,
        fps,
        apply=False,
        configuration=configuration,
        no_write_summary=True,
    )

    return command.finish() if result is True else command.fail()


async def take_fvc_loop(
    command: JaegerCommandType,
    fps: FPS,
    exposure_time: float | None = None,
    fbi_level: float | None = None,
    one: bool = False,
    max_iterations: int | None = None,
    stack: int = 3,
    plot: bool = True,
    apply: bool = True,
    max_correction: float | None = None,
    target_90_percentile: float | None = None,
    k: float | None = None,
    centroid_method: str | None = None,
    use_invkin: bool = True,
    no_write_summary: bool = False,
    configuration: BaseConfiguration | None = None,
):
    """Helper to take an FVC loop that can be called externally."""

    exposure_time = exposure_time or config["fvc"]["exposure_time"]
    fbi_level = fbi_level if fbi_level is not None else config["fvc"]["fbi_level"]
    assert isinstance(exposure_time, float) and isinstance(fbi_level, (float, int))

    configuration = configuration or fps.configuration

    if configuration is None:
        command.error("Configuration not loaded.")
        return False

    fvc = FVC(fps.observatory, command=command)

    # Check that the rotator is halted.
    if config["fvc"]["check_rotator"] is True:
        axis_cmd = await command.send_command("keys", "getFor=tcc AxisCmdState")
        if axis_cmd.status.did_fail:
            command.warning("Cannot check the status of the rotator.")
        else:
            rot_status = axis_cmd.replies.get("AxisCmdState")[2]
            if rot_status != "Halted":
                command.error(f"Cannot expose FVC while the rotator is {rot_status}.")
                return False
            else:
                command.debug("The rotator is halted.")

    command.debug("Turning LEDs on.")
    await command.send_command("jaeger", f"ieb fbi led1 led2 {fbi_level}")

    if one is True and apply is True:
        command.warning(
            "One correction will be applied. The confSummaryF "
            "file will not reflect the final state."
        )

    max_iterations = max_iterations or config["fvc"]["max_fvc_iterations"]
    target_90_percentile = target_90_percentile or config["fvc"]["target_90_percentile"]

    current_rms = None
    delta_rms = None

    filename = None
    proc_image_saved = False

    # Flag to determine when to exit the loop.
    reached: bool = False
    failed: bool = False

    try:

        n = 1
        while True:
            command.info(f"FVC iteration {n}")

            filename = None
            proc_image_saved: bool = False

            # 1. Expose the FVC
            command.debug("Taking exposure with fliswarm.")
            filename = await fvc.expose(exposure_time=exposure_time, stack=stack)
            command.debug(fvc_filename=str(filename))

            fvc.iteration = n

            # 2. Process the new image.
            positioner_coords = fps.get_positions_dict()
            await run_in_executor(
                fvc.process_fvc_image,
                filename,
                positioner_coords,
                configuration=configuration,
                plot=plot,
                centroid_method=centroid_method,
                use_new_invkin=use_invkin,
            )

            # 3. Set current RMS and delta.
            new_rms = round(fvc.fitrms * 1000.0, 2)

            command.info(fvc_rms=new_rms)

            if current_rms is None:
                pass
            else:
                delta_rms = current_rms - new_rms
                command.info(fvc_deltarms=delta_rms)
            current_rms = new_rms

            command.info(fvc_perc_90=round(fvc.perc_90 * 1000.0, 2))
            command.info(fvc_percent_reached=round(fvc.fvc_percent_reached, 1))

            # 4. Check if we have reached the distance criterion.
            if target_90_percentile and fvc.perc_90 * 1000.0 <= target_90_percentile:
                command.info("Target 90% percentile reached.")
                reached = True

            # 4. Update current positions and calculate offsets.
            command.debug("Calculating offsets.")
            await fps.update_position()
            await run_in_executor(
                fvc.calculate_offsets,
                fps.get_positions(),
                k=k,
                max_correction=max_correction,
            )

            # 5. Apply corrections.
            if reached is False and apply is True:
                if n == max_iterations and one is False:
                    command.debug("Not applying correction during the last iteration.")
                else:
                    await fvc.apply_correction()

            # 6. Save processed file.
            proc_path = filename.with_name("proc-" + filename.name)
            command.debug(f"Saving processed image {proc_path}")
            await fvc.write_proc_image(proc_path)
            proc_image_saved = True

            if reached is True:
                break

            if one is True or apply is False:
                command.warning("Cancelling FVC loop after one iteration.")
                break

            if n == max_iterations:
                command.warning("Maximum number of iterations reached.")
                break

            n += 1

    except Exception as err:
        failed = True
        command.error(error=f"Failed processing image: {err}")
        return False

    finally:
        try:
            if (
                not isinstance(fps.configuration, ManualConfiguration)
                and no_write_summary is False
                and failed is False
            ):
                command.info("Saving confSummaryF file.")
                await fvc.write_summary_F()

            if proc_image_saved is False:
                if filename is not None and fvc.proc_hdu is not None:
                    proc_path = filename.with_name("proc-" + filename.name)
                    command.debug(f"Saving processed image {proc_path}")
                    await fvc.write_proc_image(proc_path)
                else:
                    command.warning("Cannot write processed image.")
        except Exception:
            pass

        command.debug("Turning LEDs off.")
        await command.send_command("jaeger", "ieb fbi led1 led2 0")

    if reached is True or apply is False:
        return True
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-27
# @Filename: can.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from jaeger.can import JaegerCAN

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from jaeger import FPS

    from ..actor import JaegerActor

__all__ = ["can"]


@jaeger_parser.group()
def can():
    """Allows to connect/disconnect the CAN interfaces."""


@can.command()
async def connect(command: Command[JaegerActor], fps: FPS):
    """Connect the CAN interfaces."""

    if isinstance(fps.can, JaegerCAN):
        fps.can.stop()

    await fps.start_can()
    if fps.pid_lock is not None:
        fps.pid_lock.close()
        fps.pid_lock = None

    return command.finish(text="CAN has been reconnected.")


@can.command()
async def disconnect(command: Command[JaegerActor], fps: FPS):
    """Connect the CAN interfaces."""

    if isinstance(fps.can, JaegerCAN):
        fps.can.stop()

    return command.finish(text="CAN has been disconnected.")
